/// \file ijkcontourteeIO.txx
/// ijk templates defining contour tree IO routines
/// IO routines write output in graphviz .dot format
/// Version 0.1.0

/*
  IJK: Isosurface Jeneration Kode
  Copyright (C) 2009 Rephael Wenger

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  (LGPL) as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _IJKCONTOURTREEIO_
#define _IJKCONTOURTREEIO_

#include <iostream>
#include <sstream>

#include "ijkcontourtree.txx"

/// Contour tree IO functions
namespace IJK {
namespace CONTOUR_TREE {

    // **************************************************
    // Output tree
    // **************************************************

    class DOT_PARAM {

    public:
      typedef enum {DEFAULT_RANK_DIR, TOP_TO_BOTTOM, BOTTOM_TO_TOP} RANK_DIR;

      RANK_DIR rank_dir;
      bool is_directed;
      int first_label;

    public:
      DOT_PARAM(): 
	rank_dir(DEFAULT_RANK_DIR), is_directed(false), first_label(0)
      {};

      // get functions
      bool IsDirected() const { return(is_directed); };
    };

    /// Output dot header
    void output_dot_header
    (std::ostream & out, const std::string & graph_name,
     const DOT_PARAM & dot_param)
    {
      using std::endl;

      if (dot_param.is_directed) { out << "digraph "; }
      else {out << "graph "; };
      out << graph_name << " {" << endl;
    };

    /// Output dot graph attibutes
    void output_dot_graph_attributes
    (std::ostream & out, const DOT_PARAM & dot_param)
    {
      using std::endl;

      switch(dot_param.rank_dir) {

      case DOT_PARAM::BOTTOM_TO_TOP: 
	out << "  rankdir=BT;" << endl;
	break;

      case DOT_PARAM::TOP_TO_BOTTOM: 
	out << "  rankdir=TB;" << endl;
	break;
      }

    };

    /// Output dot vertex attibutes
    template <class NODE_TYPE, class NUM_TYPE, class STYPE>
    void output_dot_vertex_attributes
    (std::ostream & out, const NODE_TYPE * tree, const NUM_TYPE num_nodes, 
     const STYPE * scalar, const DOT_PARAM & dot_param)
    {
      using std::endl;
      using std::string;
      std::ostringstream label;

      const int first_label = dot_param.first_label;

      for (NUM_TYPE i = 0; i < num_nodes; i++) {
	NUM_TYPE iv = tree[i].Ident();
	STYPE s = scalar[iv];

	label.str("");
	label << iv+first_label << " (" << s << ")";

	out << "  " << iv << " [label=\"" << label.str() << "\"];" << endl;
      }
      out << endl;
    };

    /// Output dot edges
    template <class NODE_TYPE, class NUM_TYPE>
    void output_dot_contour_edges
    (std::ostream & out, const NODE_TYPE * tree, const NUM_TYPE num_nodes,
     const DOT_PARAM & dot_param)
    {
      using std:: endl;

      std::string edge_string = " -- ";
      if (dot_param.is_directed)
	{ edge_string = " -> "; }

      for (NUM_TYPE i = 0; i < num_nodes; i++) {
	if (!tree[i].IsDeleted() && !tree[i].IsRoot()) {
	  out << "  ";
	  if (tree[i].EdgeType() == JOIN_EDGE)
	    { out << i << edge_string << ParentIndex(tree, i); }
	  else
	    { out << ParentIndex(tree, i) << edge_string << i; }
	  out << ";" << endl;
	}
      }
    };

    /// Output dot edges
    template <class NODE_TYPE, class NUM_TYPE>
    void output_id_dot_contour_edges
    (std::ostream & out, const NODE_TYPE * tree, const NUM_TYPE num_nodes,
     const DOT_PARAM & dot_param)
    {
      using std:: endl;

      std::string edge_string = " -- ";
      if (dot_param.is_directed)
	{ edge_string = " -> "; }

      for (NUM_TYPE i = 0; i < num_nodes; i++) {
	if (!tree[i].IsDeleted() && !tree[i].IsRoot()) {
	  out << "  ";
	  if (tree[i].EdgeType() == JOIN_EDGE)
	    { out << tree[i].Ident() << edge_string << tree[i].Parent()->Ident(); }
	  else
	    { out << tree[i].Parent()->Ident() << edge_string << tree[i].Ident(); }
	  out << ";" << endl;
	}
      }
    };

    /// Output dot edges
    template <class NODE_TYPE, class NUM_TYPE>
    void output_id_dot_edges
    (std::ostream & out, const NODE_TYPE * tree, const NUM_TYPE num_nodes,
     const DOT_PARAM & dot_param)
    {
      using std:: endl;

      for (NUM_TYPE i = 0; i < num_nodes; i++) {
	if (!tree[i].IsDeleted() && !tree[i].IsRoot()) {
	  out << "  ";
	  if (dot_param.is_directed) {
	    out << tree[i].Ident() << " -> "
		<< tree[i].Parent()->Ident();
	  }
	  else
	    { out << tree[i].Ident() << " -- " 
		  << tree[i].Parent()->Ident(); }
	  out << ";" << endl;
	}
      }
    };

    /// Output tree in graphviz dot format
    template <class NODE_TYPE, class NUM_TYPE>
    void output_dot(std::ostream & out, const std::string & graph_name,
		    const NODE_TYPE * tree, const NUM_TYPE num_nodes,
		    const DOT_PARAM & dot_param)
    {
      using std:: endl;

      // opening {
      output_dot_header(out, graph_name, dot_param);
      output_dot_graph_attributes(out, dot_param);
      output_dot_edges(out, tree, num_nodes, dot_param);

      // closing }
      out << "}" << endl;
    }

    /// Output tree in graphviz dot format
    template <class NODE_TYPE, class NUM_TYPE>
    void output_dot_contour
    (std::ostream & out, const std::string & graph_name,
     const NODE_TYPE * tree, const NUM_TYPE num_nodes,
     const DOT_PARAM & dot_param)
    {
      using std:: endl;

      // opening {
      output_dot_header(out, graph_name, dot_param);
      output_dot_graph_attributes(out, dot_param);
      output_dot_contour_edges(out, tree, num_nodes, dot_param);

      // closing }
      out << "}" << endl;
    }

    /// Output tree identifiers in graphviz dot format
    template <class NODE_TYPE, class NUM_TYPE>
    void output_id_dot(std::ostream & out, const std::string & graph_name,
		       const NODE_TYPE * tree, const NUM_TYPE num_nodes,
		       const DOT_PARAM & dot_param)
    {
      using std:: endl;

      output_dot_header(out, graph_name, dot_param);
      output_dot_graph_attributes(out, dot_param);
      output_id_dot_edges(out, tree, num_nodes, dot_param);

      // closing }
      out << "}" << endl;
    }

    /// Output tree identifiers in graphviz dot format
    template <class NODE_TYPE, class NUM_TYPE>
    void output_id_dot_contour
    (std::ostream & out, const std::string & graph_name,
     const NODE_TYPE * tree, const NUM_TYPE num_nodes, 
     const DOT_PARAM & dot_param)
    {
      using std:: endl;

      output_dot_header(out, graph_name, dot_param);
      output_dot_graph_attributes(out, dot_param);
      output_id_dot_contour_edges(out, tree, num_nodes, dot_param);

      // closing }
      out << "}" << endl;
    }


    /// Output tree identifiers in graphviz dot format
    template <class NODE_TYPE, class NUM_TYPE, class STYPE>
    void output_id_dot_contour
    (std::ostream & out, const std::string & graph_name,
     const NODE_TYPE * tree, const NUM_TYPE num_nodes, 
     const STYPE * scalar,
     const DOT_PARAM & dot_param)
    {
      using std:: endl;

      output_dot_header(out, graph_name, dot_param);
      output_dot_graph_attributes(out, dot_param);
      output_dot_vertex_attributes(out, tree, num_nodes, scalar, dot_param);
      output_id_dot_contour_edges(out, tree, num_nodes, dot_param);

      // closing }
      out << "}" << endl;
    }

  }
}

#endif
